@@ -16,6 +16,8 @@
 //
 //
 
+#include <grpc/support/port_platform.h>
+
 #include "src/core/ext/filters/client_channel/subchannel_index.h"
 
 #include <stdbool.h>
@@ -37,58 +39,37 @@ static gpr_mu g_mu;
 static gpr_refcount g_refcount;
 
 struct grpc_subchannel_key {
-  grpc_subchannel_args args;
+  grpc_channel_args* args;
 };
 
-static bool g_force_creation = false;
-
 static grpc_subchannel_key* create_key(
-    const grpc_subchannel_args* args,
+    const grpc_channel_args* args,
     grpc_channel_args* (*copy_channel_args)(const grpc_channel_args* args)) {
   grpc_subchannel_key* k =
       static_cast<grpc_subchannel_key*>(gpr_malloc(sizeof(*k)));
-  k->args.filter_count = args->filter_count;
-  if (k->args.filter_count > 0) {
-    k->args.filters = static_cast<const grpc_channel_filter**>(
-        gpr_malloc(sizeof(*k->args.filters) * k->args.filter_count));
-    memcpy(reinterpret_cast<grpc_channel_filter*>(k->args.filters),
-           args->filters, sizeof(*k->args.filters) * k->args.filter_count);
-  } else {
-    k->args.filters = nullptr;
-  }
-  k->args.args = copy_channel_args(args->args);
+  k->args = copy_channel_args(args);
   return k;
 }
 
-grpc_subchannel_key* grpc_subchannel_key_create(
-    const grpc_subchannel_args* args) {
+grpc_subchannel_key* grpc_subchannel_key_create(const grpc_channel_args* args) {
   return create_key(args, grpc_channel_args_normalize);
 }
 
 static grpc_subchannel_key* subchannel_key_copy(grpc_subchannel_key* k) {
-  return create_key(&k->args, grpc_channel_args_copy);
+  return create_key(k->args, grpc_channel_args_copy);
 }
 
 int grpc_subchannel_key_compare(const grpc_subchannel_key* a,
                                 const grpc_subchannel_key* b) {
-  if (g_force_creation) return false;
-  int c = GPR_ICMP(a->args.filter_count, b->args.filter_count);
-  if (c != 0) return c;
-  if (a->args.filter_count > 0) {
-    c = memcmp(a->args.filters, b->args.filters,
-               a->args.filter_count * sizeof(*a->args.filters));
-    if (c != 0) return c;
-  }
-  return grpc_channel_args_compare(a->args.args, b->args.args);
+  return grpc_channel_args_compare(a->args, b->args);
 }
 
 void grpc_subchannel_key_destroy(grpc_subchannel_key* k) {
-  gpr_free(reinterpret_cast<grpc_channel_args*>(k->args.filters));
-  grpc_channel_args_destroy(const_cast<grpc_channel_args*>(k->args.args));
+  grpc_channel_args_destroy(k->args);
   gpr_free(k);
 }
 
-static void sck_avl_destroy(void* p, void* user_data) {
+static void sck_avl_destroy(void* p, void* unused) {
   grpc_subchannel_key_destroy(static_cast<grpc_subchannel_key*>(p));
 }
 
@@ -101,7 +82,7 @@ static long sck_avl_compare(void* a, void* b, void* unused) {
                                      static_cast<grpc_subchannel_key*>(b));
 }
 
-static void scv_avl_destroy(void* p, void* user_data) {
+static void scv_avl_destroy(void* p, void* unused) {
   GRPC_SUBCHANNEL_WEAK_UNREF((grpc_subchannel*)p, "subchannel_index");
 }
 
@@ -134,7 +115,7 @@ void grpc_subchannel_index_shutdown(void) {
 void grpc_subchannel_index_unref(void) {
   if (gpr_unref(&g_refcount)) {
     gpr_mu_destroy(&g_mu);
-    grpc_avl_unref(g_subchannel_index, grpc_core::ExecCtx::Get());
+    grpc_avl_unref(g_subchannel_index, nullptr);
   }
 }
 
@@ -144,13 +125,12 @@ grpc_subchannel* grpc_subchannel_index_find(grpc_subchannel_key* key) {
   // Lock, and take a reference to the subchannel index.
   // We don't need to do the search under a lock as avl's are immutable.
   gpr_mu_lock(&g_mu);
-  grpc_avl index = grpc_avl_ref(g_subchannel_index, grpc_core::ExecCtx::Get());
+  grpc_avl index = grpc_avl_ref(g_subchannel_index, nullptr);
   gpr_mu_unlock(&g_mu);
 
   grpc_subchannel* c = GRPC_SUBCHANNEL_REF_FROM_WEAK_REF(
-      (grpc_subchannel*)grpc_avl_get(index, key, grpc_core::ExecCtx::Get()),
-      "index_find");
-  grpc_avl_unref(index, grpc_core::ExecCtx::Get());
+      (grpc_subchannel*)grpc_avl_get(index, key, nullptr), "index_find");
+  grpc_avl_unref(index, nullptr);
 
   return c;
 }
@@ -166,13 +146,11 @@ grpc_subchannel* grpc_subchannel_index_register(grpc_subchannel_key* key,
     // Compare and swap loop:
     // - take a reference to the current index
     gpr_mu_lock(&g_mu);
-    grpc_avl index =
-        grpc_avl_ref(g_subchannel_index, grpc_core::ExecCtx::Get());
+    grpc_avl index = grpc_avl_ref(g_subchannel_index, nullptr);
     gpr_mu_unlock(&g_mu);
 
     // - Check to see if a subchannel already exists
-    c = static_cast<grpc_subchannel*>(
-        grpc_avl_get(index, key, grpc_core::ExecCtx::Get()));
+    c = static_cast<grpc_subchannel*>(grpc_avl_get(index, key, nullptr));
     if (c != nullptr) {
       c = GRPC_SUBCHANNEL_REF_FROM_WEAK_REF(c, "index_register");
     }
@@ -181,11 +159,9 @@ grpc_subchannel* grpc_subchannel_index_register(grpc_subchannel_key* key,
       need_to_unref_constructed = true;
     } else {
       // no -> update the avl and compare/swap
-      grpc_avl updated =
-          grpc_avl_add(grpc_avl_ref(index, grpc_core::ExecCtx::Get()),
-                       subchannel_key_copy(key),
-                       GRPC_SUBCHANNEL_WEAK_REF(constructed, "index_register"),
-                       grpc_core::ExecCtx::Get());
+      grpc_avl updated = grpc_avl_add(
+          grpc_avl_ref(index, nullptr), subchannel_key_copy(key),
+          GRPC_SUBCHANNEL_WEAK_REF(constructed, "index_register"), nullptr);
 
       // it may happen (but it's expected to be unlikely)
       // that some other thread has changed the index:
@@ -197,9 +173,9 @@ grpc_subchannel* grpc_subchannel_index_register(grpc_subchannel_key* key,
       }
       gpr_mu_unlock(&g_mu);
 
-      grpc_avl_unref(updated, grpc_core::ExecCtx::Get());
+      grpc_avl_unref(updated, nullptr);
     }
-    grpc_avl_unref(index, grpc_core::ExecCtx::Get());
+    grpc_avl_unref(index, nullptr);
   }
 
   if (need_to_unref_constructed) {
@@ -216,24 +192,22 @@ void grpc_subchannel_index_unregister(grpc_subchannel_key* key,
     // Compare and swap loop:
     // - take a reference to the current index
     gpr_mu_lock(&g_mu);
-    grpc_avl index =
-        grpc_avl_ref(g_subchannel_index, grpc_core::ExecCtx::Get());
+    grpc_avl index = grpc_avl_ref(g_subchannel_index, nullptr);
     gpr_mu_unlock(&g_mu);
 
     // Check to see if this key still refers to the previously
     // registered subchannel
-    grpc_subchannel* c = static_cast<grpc_subchannel*>(
-        grpc_avl_get(index, key, grpc_core::ExecCtx::Get()));
+    grpc_subchannel* c =
+        static_cast<grpc_subchannel*>(grpc_avl_get(index, key, nullptr));
     if (c != constructed) {
-      grpc_avl_unref(index, grpc_core::ExecCtx::Get());
+      grpc_avl_unref(index, nullptr);
       break;
     }
 
     // compare and swap the update (some other thread may have
     // mutated the index behind us)
     grpc_avl updated =
-        grpc_avl_remove(grpc_avl_ref(index, grpc_core::ExecCtx::Get()), key,
-                        grpc_core::ExecCtx::Get());
+        grpc_avl_remove(grpc_avl_ref(index, nullptr), key, nullptr);
 
     gpr_mu_lock(&g_mu);
     if (index.root == g_subchannel_index.root) {
@@ -242,11 +216,7 @@ void grpc_subchannel_index_unregister(grpc_subchannel_key* key,
     }
     gpr_mu_unlock(&g_mu);
 
-    grpc_avl_unref(updated, grpc_core::ExecCtx::Get());
-    grpc_avl_unref(index, grpc_core::ExecCtx::Get());
+    grpc_avl_unref(updated, nullptr);
+    grpc_avl_unref(index, nullptr);
   }
 }
-
-void grpc_subchannel_index_test_only_set_force_creation(bool force_creation) {
-  g_force_creation = force_creation;
-}
