@@ -130,10 +130,11 @@ void grpc_init(void) {
     grpc_channel_init_init();
     grpc_core::channelz::ChannelzRegistry::Init();
     grpc_security_pre_init();
+    grpc_core::ApplicationCallbackExecCtx::GlobalInit();
     grpc_core::ExecCtx::GlobalInit();
     grpc_iomgr_init();
     gpr_timers_global_init();
-    grpc_handshaker_factory_registry_init();
+    grpc_core::HandshakerRegistry::Init();
     grpc_security_init();
     for (i = 0; i < g_number_of_plugins; i++) {
       if (g_all_of_the_plugins[i].init != nullptr) {
@@ -161,10 +162,11 @@ void grpc_shutdown(void) {
   if (--g_initializations == 0) {
     {
       grpc_core::ExecCtx exec_ctx(0);
+      grpc_iomgr_shutdown_background_closure();
       {
         grpc_timer_manager_set_threading(
             false);  // shutdown timer_manager thread
-        grpc_executor_shutdown();
+        grpc_core::Executor::ShutdownAll();
         for (i = g_number_of_plugins; i >= 0; i--) {
           if (g_all_of_the_plugins[i].destroy != nullptr) {
             g_all_of_the_plugins[i].destroy();
@@ -175,13 +177,14 @@ void grpc_shutdown(void) {
       gpr_timers_global_destroy();
       grpc_tracer_shutdown();
       grpc_mdctx_global_shutdown();
-      grpc_handshaker_factory_registry_shutdown();
+      grpc_core::HandshakerRegistry::Shutdown();
       grpc_slice_intern_shutdown();
       grpc_core::channelz::ChannelzRegistry::Shutdown();
       grpc_stats_shutdown();
       grpc_core::Fork::GlobalShutdown();
     }
     grpc_core::ExecCtx::GlobalShutdown();
+    grpc_core::ApplicationCallbackExecCtx::GlobalShutdown();
   }
   gpr_mu_unlock(&g_init_mu);
 }
